using UnityEngine;
using UnityEngine.XR;
using UnityEngine.Rendering;

namespace moyashi {
  public struct CSMeta {
    public int kernelIndex;
    public uint threadGroupsX;
    public uint threadGroupsY;
    public uint threadGroupsZ;
  }

  public struct CommandBufferResources {
    public bool ready;
    public Material prepassMaterial;
    public Material colorCorrectionMaterial;
    public ComputeShader compute;
    public ComputeShader lensCompute;
    public CSMeta computeMeta;
    public CSMeta lensComputeMeta;
  }

  class CommandBufferBuilder {
    public static CommandBuffer buildFullCommandBuffer(CommandBufferResources resources, int width, int height, bool isVR) {
      if (isVR) {
        return buildFullVR(resources, width, height);
      } else {
        return buildFullDesktop(resources, width, height);
      }
    }

    public static CommandBuffer buildFullDesktop(CommandBufferResources resources, int width, int height) {
      CommandBuffer commandBuffer = new CommandBuffer();
      commandBuffer.name = "Moyashi post processing (Desktop)";

      // TODO: setting MSAA here to 2 does seem to look better, but breaks the compute shaders

      int pingRT = Shader.PropertyToID("_MoyashiPostProcessingPingRT");
      commandBuffer.GetTemporaryRT(pingRT, -1, -1, 0, FilterMode.Point, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 1, true);

      int pongRT = Shader.PropertyToID("_MoyashiPostProcessingPongRT");
      commandBuffer.GetTemporaryRT(pongRT, -1, -1, 0, FilterMode.Point, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 1, true);

      commandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, pongRT, resources.prepassMaterial);
      commandBuffer.Blit(pongRT, pingRT, resources.colorCorrectionMaterial);

      // commandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, pingRT, this.colorCorrectionMaterial);
      // commandBuffer.CopyTexture(BuiltinRenderTextureType.CameraTarget, pingRT);

      commandBuffer.SetComputeTextureParam(resources.compute, resources.computeMeta.kernelIndex, "Read", pingRT);
      commandBuffer.SetComputeTextureParam(resources.compute, resources.computeMeta.kernelIndex, "Result", pongRT);
      commandBuffer.DispatchCompute(resources.compute, resources.computeMeta.kernelIndex, 
        width/(int)resources.computeMeta.threadGroupsX, 
        height/(int)resources.computeMeta.threadGroupsY, 
        (int)resources.computeMeta.threadGroupsZ
      );

      commandBuffer.SetComputeTextureParam(resources.lensCompute, resources.lensComputeMeta.kernelIndex, "Read", pongRT);
      commandBuffer.SetComputeTextureParam(resources.lensCompute, resources.lensComputeMeta.kernelIndex, "Result", pingRT);
      commandBuffer.DispatchCompute(resources.lensCompute, resources.lensComputeMeta.kernelIndex, 
        width/(int)resources.lensComputeMeta.threadGroupsX, 
        height/(int)resources.lensComputeMeta.threadGroupsY, 
        (int)resources.lensComputeMeta.threadGroupsZ
      );

      commandBuffer.Blit(pingRT, BuiltinRenderTextureType.CameraTarget);
      // commandBuffer.Blit(renderTexture, BuiltinRenderTextureType.None);
      // commandBuffer.Blit(pingRT, BuiltinRenderTextureType.CameraTarget);
      // commandBuffer.CopyTexture(pingRT, BuiltinRenderTextureType.CameraTarget);
      // commandBuffer.Blit(renderTexture, null as RenderTexture);
      commandBuffer.ReleaseTemporaryRT(pingRT);
      commandBuffer.ReleaseTemporaryRT(pongRT);

      return commandBuffer;
    }

    public static CommandBuffer buildFullVR(CommandBufferResources resources, int width, int height) {
      CommandBuffer commandBuffer = new CommandBuffer();
      commandBuffer.name = "Moyashi post processing (VR)";

      int VRRT = Shader.PropertyToID("_MoyashiPostProcessingVRRT");

      // RenderTextureDescriptor rtDesc = new RenderTextureDescriptor(width, height);

      RenderTextureDescriptor rtDesc = XRSettings.eyeTextureDesc;

      rtDesc.autoGenerateMips = false;
      // rtDesc.bindMS = true; // msaaSamples ? TODO
      rtDesc.colorFormat = RenderTextureFormat.Default;
      rtDesc.depthBufferBits = 0;
      rtDesc.dimension = TextureDimension.Tex2D;
      // rtDesc.enableRandomWrite = false; // enable this if compute shader end up being used here
      // rtDesc.memoryless = RenderTextureMemoryless.None;
      // rtDesc.msaaSamples = 1; // TODO
      rtDesc.sRGB = true;
      rtDesc.useMipMap = false;
      // rtDesc.vrUsage = VRTextureUsage.TwoEyes;

      // rtDesc.shadowSamplingMode = // not needed i think
      // rtDesc.volumeDepth // not needed
      // rtDesc.flags =  // is this needed if filling out other shit?

      commandBuffer.GetTemporaryRT(VRRT, rtDesc, FilterMode.Trilinear);
      // commandBuffer.GetTemporaryRT(VRRT, -1, -1, 0, FilterMode.Point, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 4, false);

      commandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, VRRT, resources.prepassMaterial);
      commandBuffer.Blit(VRRT, BuiltinRenderTextureType.CameraTarget, resources.colorCorrectionMaterial);

      commandBuffer.ReleaseTemporaryRT(VRRT);

      return commandBuffer;
    }

    // this is just for screenshot + thumbnail camera
    public static CommandBuffer buildLightWeightCommandBuffer(CommandBufferResources resources, int width, int height) {
      CommandBuffer commandBuffer = new CommandBuffer();
      commandBuffer.name = "Moyashi lightweight post processing";

      int LWRT = Shader.PropertyToID("_MoyashiPostProcessingLWRT");

      commandBuffer.GetTemporaryRT(LWRT, width, height, 0, FilterMode.Point, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 1, false);

      commandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, LWRT, resources.prepassMaterial);
      commandBuffer.Blit(LWRT, BuiltinRenderTextureType.CameraTarget, resources.colorCorrectionMaterial);

      commandBuffer.ReleaseTemporaryRT(LWRT);

      return commandBuffer;
    }
  }
}