using System.Linq;

using UnityEngine;
using UnityEngine.PostProcessing;
using UnityEngine.Rendering;

namespace moyashi {

  public class EffectsManager : MonoBehaviour {
    private static int DEFAULT_CULLING_MASK = -142606337;

    private int width = 0;
    private int height = 0;

    private Camera postProcessingCamera;
    private Camera mainCamera;
    private Camera mainWindowCamera;
    private int uiMask = 0;
    private int worldMask = 0;
    
    private float originalMainDepth = 0.0f;
  
    private BloomModel.Settings bloomSettings;
    private DepthOfFieldModel.Settings depthOfFieldSettings;

    private GameObject autoFocusTarget = null;

    private PostProcessingBehaviour postProcessingBehaviour;
    public PostProcessingProfile profile { get; private set; }

    private CommandBufferResources commandBufferResources;
    private CommandBuffer postProcessingCameraCommandBuffer;
    private CommandBuffer mainWindowCameraCommandBuffer;
    private CommandBuffer screenshotCameraCommandBuffer;
    private CommandBuffer hiResScreenshotCameraCommandBuffer;

    // copy to main cam
    private CommandBuffer finalCommandBuffer;

    private RenderTexture renderTexture;

    // https://docs.unity3d.com/2018.1/Documentation/ScriptReference/Rendering.CameraEvent.html
    private CameraEvent PRE_PROC_EVENT = CameraEvent.BeforeForwardAlpha;
    private CameraEvent POST_PROC_EVENT = CameraEvent.BeforeImageEffects; // BeforeImageEffects // why doesnt AfterImageEffects work when using render texture?

    // public void OnPreRender() {
    //   postProcessingCamera.targetTexture = renderTexture;
    // }

    // public void OnPostRender() {
    //   postProcessingCamera.targetTexture = null;
    //   // Graphics.Blit(renderTexture, null as RenderTexture); // ??????????? why isnt this needed
    //   // postProcessingCamera.targetTexture = renderTexture;
    // }

    // public void setPrepassShader(Shader shader) {
    //   // getOrAddCustomPostProcessingPrepassComponent().setPrepassShader(shader);
    //   prepassShader = shader;
    //   prepassMaterial = new Material(shader);

    //   prepassCommandBuffer = new CommandBuffer();
    //   prepassCommandBuffer.name = "Moyashi post processing prepass";

    //   int screenCopyId = Shader.PropertyToID("_MoyashiPrePassRT");
    //   // SuperController.LogMessage($"id: {screenCopyId}");
    //   prepassCommandBuffer.GetTemporaryRT(screenCopyId, -1, -1, 0, FilterMode.Bilinear, RenderTextureFormat.Default, RenderTextureReadWrite.sRGB, 1, true);
    //   // prepassCommandBuffer.SetRenderTarget(BuiltinRenderTextureType.CameraTarget);
    //   prepassCommandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, screenCopyId, this.prepassMaterial);
    //   prepassCommandBuffer.Blit(screenCopyId, BuiltinRenderTextureType.CameraTarget);

    //   prepassCommandBuffer.ReleaseTemporaryRT(screenCopyId);
    //   // postProcessingCamera.AddCommandBuffer(PRE_PROC_EVENT, prepassCommandBuffer);
    //   // SuperController.LogMessage("set command buffer");

    //   updateUniforms();
    // }

    // TODO: completely avoid blit to backbuffer? https://forum.unity.com/threads/are-command-buffers-quicker-than-onrenderimage.470639/
    public void setShaders(Shader prepassShader, Shader colorCorrectionShader, ComputeShader compute, ComputeShader lensCompute) {
      commandBufferResources.prepassMaterial = new Material(prepassShader);
      commandBufferResources.colorCorrectionMaterial = new Material(colorCorrectionShader);
      commandBufferResources.compute = compute;
      commandBufferResources.lensCompute = lensCompute;

      commandBufferResources.computeMeta.kernelIndex = commandBufferResources.compute.FindKernel("MoyashiCompute");

      commandBufferResources.compute.GetKernelThreadGroupSizes(commandBufferResources.computeMeta.kernelIndex,
        out commandBufferResources.computeMeta.threadGroupsX,
        out commandBufferResources.computeMeta.threadGroupsY,
        out commandBufferResources.computeMeta.threadGroupsZ
      );

      commandBufferResources.lensComputeMeta.kernelIndex = commandBufferResources.lensCompute.FindKernel("MoyashiLens");

      commandBufferResources.lensCompute.GetKernelThreadGroupSizes(commandBufferResources.lensComputeMeta.kernelIndex,
        out commandBufferResources.lensComputeMeta.threadGroupsX,
        out commandBufferResources.lensComputeMeta.threadGroupsY,
        out commandBufferResources.lensComputeMeta.threadGroupsZ
      );

      updateUniforms();

      commandBufferResources.ready = true;

      postProcessingCameraCommandBuffer = CommandBufferBuilder.buildFullCommandBuffer(commandBufferResources, width, height, postProcessingCamera.stereoEnabled);
      postProcessingCamera.AddCommandBuffer(POST_PROC_EVENT, postProcessingCameraCommandBuffer);

      // while this works, it only works if the window camera is turned on beforehand... so kinda annoying
      // Object.FindObjectsOfType<Camera>().ToList().ForEach((c) => SuperController.LogMessage(c.name));
      // mainWindowCamera = Object.FindObjectsOfType<Camera>().First((c) => c.name == "MainWindowCamera");
      // if (mainWindowCamera != null) {
      //   mainWindowCameraCommandBuffer = CommandBufferBuilder.buildFullCommandBuffer(commandBufferResources, width, height, mainWindowCamera.stereoEnabled);
      //   mainWindowCamera.AddCommandBuffer(POST_PROC_EVENT, mainWindowCameraCommandBuffer);
      // }

      {
        var cam = SuperController.singleton.screenshotCamera;
        var camWidth = cam.pixelWidth;
        var camHeight = cam.pixelHeight;

        screenshotCameraCommandBuffer = CommandBufferBuilder.buildLightWeightCommandBuffer(commandBufferResources, camWidth, camHeight);
        cam.AddCommandBuffer(POST_PROC_EVENT, screenshotCameraCommandBuffer);
      }

      {
        var cam = SuperController.singleton.hiResScreenshotCamera;
        var camWidth = cam.pixelWidth;
        var camHeight = cam.pixelHeight;

        hiResScreenshotCameraCommandBuffer = CommandBufferBuilder.buildLightWeightCommandBuffer(commandBufferResources, camWidth, camHeight);
        cam.AddCommandBuffer(POST_PROC_EVENT, hiResScreenshotCameraCommandBuffer);
      }
    }

    public void setBloomEnabled(bool b) {
      profile.bloom.enabled = b;
      postProcessingBehaviour.enabled = profile.bloom.enabled || profile.depthOfField.enabled;
    }

    public void setBloomThreshold(float f) {
      bloomSettings.bloom.threshold = f;
      profile.bloom.settings = bloomSettings;
    }

    public void setBloomIntensity(float f) {
      bloomSettings.bloom.intensity = f;
      profile.bloom.settings = bloomSettings;
    }

    public void setBloomRadius(float f) {
      bloomSettings.bloom.radius = f;
      profile.bloom.settings = bloomSettings;
    }

    public void setDepthOfFieldEnabled(bool b) {
      if (!postProcessingCamera.stereoEnabled) {
        profile.depthOfField.enabled = b;
        profile.depthOfField.settings = depthOfFieldSettings;
        postProcessingBehaviour.enabled = profile.bloom.enabled || profile.depthOfField.enabled;
      }
    }

    public void setDepthOfFieldFocusDistance(float f) {
      depthOfFieldSettings.focusDistance = f;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setDepthOfFieldAperture(float f) {
      depthOfFieldSettings.aperture = f;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setDepthOfFieldFocalLength(float f) {
      depthOfFieldSettings.focalLength = f;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize e) {
      depthOfFieldSettings.kernelSize = e;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setAutoFocusTarget(GameObject go) {
      autoFocusTarget = go;
    }

    public void setAtmosphereDensity(float v) {
      SharedUniforms.singleton.AtmosphereDensity = v;
      updateUniforms();
    }

    public void setAtmosphereSeaLevel(float v) {
      SharedUniforms.singleton.AtmosphereSeaLevel = v;
      updateUniforms();
    }

    public void setAtmosphereSeaLevelDensity(float v) {
      SharedUniforms.singleton.AtmosphereSeaLevelDensity = v;
      updateUniforms();
    }

    public void setAtmosphereSkyColor(Color c) {
      SharedUniforms.singleton.AtmosphereSkyColor = c;
      updateUniforms();
    }

    public void setAtmosphereSunColor(Color c) {
      SharedUniforms.singleton.AtmosphereSunColor = c;
      updateUniforms();
    }

    public void setLUTTexture(Texture2D t) {
      SharedUniforms.singleton.LUTTex = t;
      updateUniforms();
    }

    public void setChannelMixR(Vector3 c) {
      if (c != null) {
        SharedUniforms.singleton.ChannelMixR = c;
        updateUniforms();
      }      
    }

    public void setChannelMixG(Vector3 c) {
      if (c != null) {
        SharedUniforms.singleton.ChannelMixG = c;
        updateUniforms();
      }    
    }

    public void setChannelMixB(Vector3 c) {
      if (c != null) {
        SharedUniforms.singleton.ChannelMixB = c;
        updateUniforms();
      }    
    }

    public void setContrastScale(float v) {
      SharedUniforms.singleton.ContrastScale = v;
      updateUniforms();
    }

    public void setTonemappingScale(float v) {
      SharedUniforms.singleton.TonemappingScale = v;
      updateUniforms();
    }

    public void setTonemappingExposure(float v) {
      SharedUniforms.singleton.TonemappingExposure = v;
      updateUniforms();
    }

    public void setSaturationScale(float v) {
      SharedUniforms.singleton.SaturationScale = v;
      updateUniforms();
    }

    public void setSoftenScale(float v) {
      SharedUniforms.singleton.SoftenScale = v;
      updateUniforms();
    }

    public void setLUTContribution(float v) {
      SharedUniforms.singleton.LUTScale = v;
      updateUniforms();
    }

    public void setSharpeningScale(float v) {
      SharedUniforms.singleton.SharpeningScale = v;
      updateUniforms();
    }

    public void setVignetteScale(float v) {
      SharedUniforms.singleton.VignetteScale = v;
      updateUniforms();
    }

    public void setChromaticAberrationScale(float v) {
      SharedUniforms.singleton.ChromaticAberrationScale = v;
      updateUniforms();
    }

    public void updateUniforms() {
      var uniforms = SharedUniforms.singleton;

      if (commandBufferResources.prepassMaterial != null) {
        commandBufferResources.prepassMaterial.SetFloat("_AtmosphereDensity", uniforms.AtmosphereDensity);
        commandBufferResources.prepassMaterial.SetFloat("_AtmosphereSeaLevel", uniforms.AtmosphereSeaLevel);
        commandBufferResources.prepassMaterial.SetFloat("_AtmosphereSeaLevelDensity", uniforms.AtmosphereSeaLevelDensity);
        commandBufferResources.prepassMaterial.SetColor("_AtmosphereSkyColor", uniforms.AtmosphereSkyColor);
        commandBufferResources.prepassMaterial.SetColor("_AtmosphereSunColor", uniforms.AtmosphereSunColor);
      }

      if (commandBufferResources.colorCorrectionMaterial != null) {
        commandBufferResources.colorCorrectionMaterial.SetVector("_ChannelMixR", uniforms.ChannelMixR);
        commandBufferResources.colorCorrectionMaterial.SetVector("_ChannelMixG", uniforms.ChannelMixG);
        commandBufferResources.colorCorrectionMaterial.SetVector("_ChannelMixB", uniforms.ChannelMixB);

        commandBufferResources.colorCorrectionMaterial.SetFloat("_TonemappingScale", uniforms.TonemappingScale);
        commandBufferResources.colorCorrectionMaterial.SetFloat("_TonemappingExposure", uniforms.TonemappingExposure);
        commandBufferResources.colorCorrectionMaterial.SetFloat("_ContrastScale", uniforms.ContrastScale);
        commandBufferResources.colorCorrectionMaterial.SetFloat("_SaturationScale", uniforms.SaturationScale);
        commandBufferResources.colorCorrectionMaterial.SetFloat("_SoftenScale", uniforms.SoftenScale);
        commandBufferResources.colorCorrectionMaterial.SetFloat("_LUTScale", uniforms.LUTScale);

        if (uniforms.LUTTex != null) commandBufferResources.colorCorrectionMaterial.SetTexture("_LUTTex", uniforms.LUTTex);
      }

      if (commandBufferResources.compute != null) {
        recalculateSharpeningConsts();
        commandBufferResources.compute.SetFloat("_SharpeningScale", uniforms.SharpeningScale);
        commandBufferResources.compute.SetVector("_SharpeningConst0", uniforms.SharpeningConst0);
        commandBufferResources.compute.SetVector("_SharpeningConst1", uniforms.SharpeningConst1);
        
      }

      if (commandBufferResources.lensCompute != null) {
        commandBufferResources.lensCompute.SetFloat("_VignetteScale", uniforms.VignetteScale);
        commandBufferResources.lensCompute.SetFloat("_ChromaticAberrationScale", uniforms.ChromaticAberrationScale);
      }
    }

    private void OnEnable() {       
      // SuperController.LogMessage("EffectsCamera.OnEnable running");

      uiMask = 0 
             | 1 << LayerMask.NameToLayer("UI")
             | 1 << LayerMask.NameToLayer("LoadUI")
             | 1 << LayerMask.NameToLayer("ScreenUI")
             | 1 << LayerMask.NameToLayer("GUI")
             | 1 << LayerMask.NameToLayer("ControlColliders") // this is the "toggle targets" ui
      ;
      
      // worldMask = 0
      //           | 1 << LayerMask.NameToLayer("Default") // person models etc
      //           | 1 << LayerMask.NameToLayer("DefaultAndPlayerBlock") // this seems to be some CUAs?
      //           | 1 << LayerMask.NameToLayer("Water")
      //           | 1 << LayerMask.NameToLayer("Glow")
      //           | 1 << LayerMask.NameToLayer("TransparentFX")
      //           | 1 << LayerMask.NameToLayer("Select") // why is water in this layer too?????????

                // | 1 << LayerMask.NameToLayer("ControlColliders") // this is the "toggle targets" ui
                // | 1 << LayerMask.NameToLayer("PlayerBlock") // no idea
                // | 1 << LayerMask.NameToLayer("DefaultAndPlayerBlock") // no idea
                // | 1 << LayerMask.NameToLayer("Tongue")
                // | 1 << LayerMask.NameToLayer("TongueOnly")
                // | 1 << LayerMask.NameToLayer("PhysicsMesh1")
                // | 1 << LayerMask.NameToLayer("PhysicsMesh2")
                // | 1 << LayerMask.NameToLayer("WaterSurface")
                // | 1 << LayerMask.NameToLayer("Invisible")
                // | 1 << LayerMask.NameToLayer("LookAtTrigger")
                // | 1 << LayerMask.NameToLayer("Glow") // this displays selectable "glowable" things, so i guess render stuff to glow layer to make it glow
                // | 1 << LayerMask.NameToLayer("LoadUI")
                // | 1 << LayerMask.NameToLayer("ScreenUI")
                // | 1 << LayerMask.NameToLayer("GUI")
                // | 1 << LayerMask.NameToLayer("Ignore CharacterCore")
                // | 1 << LayerMask.NameToLayer("CharacterExtremities")
                // | 1 << LayerMask.NameToLayer("Player")
                // | 1 << LayerMask.NameToLayer("1Hair")
                // | 1 << LayerMask.NameToLayer("Character")
                // | 1 << LayerMask.NameToLayer("Ignore Collision")
                // | 1 << LayerMask.NameToLayer("Ignore Character")

                // | 1 << LayerMask.NameToLayer("PhysicsMesh1")
                // | 1 << LayerMask.NameToLayer("PhysicsMesh2")
                // | 1 << LayerMask.NameToLayer("PlayerBlock")
                // | 1 << LayerMask.NameToLayer("UI")
      ;

      // apparently some cuas just use random layers so... just render all except known UI ones
      worldMask = DEFAULT_CULLING_MASK & ~uiMask;

      // for (int i = 0; i < 32; i++) {
      //   SuperController.LogMessage($"layer {i} - {LayerMask.LayerToName(i)}");
      // }

      mainCamera = CameraTarget.centerTarget.targetCamera;
      originalMainDepth = mainCamera.depth;
      mainCamera.cullingMask = uiMask;

      GameObject cameraContainer = GameObject.Find("Moyashi_PostProcessingCamera");
      if (cameraContainer == null) {
        cameraContainer = new GameObject("Moyashi_PostProcessingCamera");
      }

      postProcessingCamera = cameraContainer.GetComponent<Camera>();

      if (postProcessingCamera == null) {
        postProcessingCamera = cameraContainer.AddComponent<Camera>();
      }
      
      // TODO: is this necessary
      postProcessingCamera.CopyFrom(mainCamera);

      // works perfectly in vr (in conjunction with setting culling matrix)
      if (postProcessingCamera.stereoEnabled) {
        postProcessingCamera.transform.parent = SuperController.singleton.heightAdjustTransform;
      } else {
        postProcessingCamera.transform.parent = mainCamera.transform;
      }

      // renderTexture = new RenderTexture(postProcessingCamera.pixelWidth, postProcessingCamera.pixelHeight, 24);
      // renderTexture = new RenderTexture(BuiltinRenderTextureType.CameraTarget);

      // definitely needed
      postProcessingCamera.cullingMask = worldMask;
      postProcessingCamera.depth = mainCamera.depth - 1;
      postProcessingCamera.enabled = true;
      postProcessingCamera.allowHDR = false; // TODO: was this on or off?
      // think this is necessary for the screen space atmospherics... otherwise it relys on user having at least 1 directional light
      postProcessingCamera.depthTextureMode = DepthTextureMode.Depth; // this does have a fairly big performance hit
      // postProcessingCamera.depthTextureMode = DepthTextureMode.None;

      // wtf were we rendering to before?
      postProcessingCamera.targetTexture = null; // UNITY/VaM WTF
      // postProcessingCamera.clearFlags = CameraClearFlags.Nothing;
      // postProcessingCamera.targetTexture = renderTexture;

      // postProcessingCamera.forceIntoRenderTexture = true;
      // postProcessingCamera.renderingPath = RenderingPath.DeferredShading;

      // these maybe not ncessary after all
      // postProcessingCamera.clearFlags = CameraClearFlags.Depth;
      // postProcessingCamera.forceIntoRenderTexture = true;
      // postProcessingCamera.clearStencilAfterLightingPass = true;
      // postProcessingCamera.renderingPath = RenderingPath.Forward;

      // these were unused, maybe worth messing around with at some point (especially when trying to get AO working)
      // postProcessingCamera.renderingPath = RenderingPath.VertexLit;
      // postProcessingCamera.renderingPath = RenderingPath.DeferredLighting;
      // postProcessingCamera.clearFlags = CameraClearFlags.Nothing;
      // postProcessingCamera.clearFlags = CameraClearFlags.SolidColor;
      // postProcessingCamera.depth = mainCamera.depth;
      // postProcessingCamera.depthTextureMode = DepthTextureMode.None;
      // postProcessingCamera.depthTextureMode = DepthTextureMode.DepthNormals;

      // mainCamera.targetTexture = null;
      mainCamera.clearFlags = CameraClearFlags.Nothing;

      if (true) {
        // try to get the post processing behaviour of the postProcessingCamera
        postProcessingBehaviour = postProcessingCamera.GetComponent<PostProcessingBehaviour>();

        // if it doesnt exist, create it
        if (postProcessingBehaviour == null) {
          postProcessingBehaviour = postProcessingCamera.gameObject.AddComponent<PostProcessingBehaviour>();
        } 

        profile = postProcessingBehaviour.profile;

        // i don't know if this can ever happen?
        if (profile == null) {
          postProcessingBehaviour.profile = profile = ScriptableObject.CreateInstance<PostProcessingProfile>(); 
        } 

        postProcessingBehaviour.enabled = true; // lose about 10 fps here

        // bloom, TODO: replace with MKGlow if I can figure out the memory leak, then dont need unity
        // post processor at all
        bloomSettings = BloomModel.Settings.defaultSettings;
        bloomSettings.bloom.intensity = 0.2f;
        bloomSettings.bloom.threshold = 0.75f;
        bloomSettings.bloom.softKnee = 0.25f;
        bloomSettings.bloom.radius = 3.0f;
        bloomSettings.bloom.antiFlicker = true;

        profile.bloom.enabled = true;
        profile.bloom.settings = bloomSettings;

        // temporary DoF
        if (!postProcessingCamera.stereoEnabled) {
          depthOfFieldSettings = DepthOfFieldModel.Settings.defaultSettings;
          depthOfFieldSettings.focusDistance = 1.5f;
          depthOfFieldSettings.aperture = 5.6f;
          depthOfFieldSettings.focalLength = 50.0f;
          depthOfFieldSettings.kernelSize = DepthOfFieldModel.KernelSize.Medium;

          profile.depthOfField.enabled = false;
          // profile.depthOfField.settings = depthOfFieldSettings;
        }

        if (false) {
          var debugSettings = BuiltinDebugViewsModel.Settings.defaultSettings;
          debugSettings.mode = BuiltinDebugViewsModel.Mode.MotionVectors;

          profile.debugViews.enabled = true;
          profile.debugViews.settings = debugSettings;
        }
      }

      // finalCommandBuffer = new CommandBuffer();
      // finalCommandBuffer.name = "Moyashi copy to camera";
      // finalCommandBuffer.Blit(renderTexture, BuiltinRenderTextureType.CameraTarget);
      // mainCamera.AddCommandBuffer(CameraEvent.AfterImageEffects, finalCommandBuffer);
    }

    private void Update() {
      if (!postProcessingCamera.stereoEnabled) {
        postProcessingCamera.fieldOfView = mainCamera.fieldOfView;

        if (autoFocusTarget) {
          var a = depthOfFieldSettings.focusDistance;
          var b = Vector3.Distance(mainCamera.transform.position, autoFocusTarget.transform.position);
          setDepthOfFieldFocusDistance(Mathf.Lerp(a, b, 5.0f * Time.deltaTime));
        }
      }

      if (Screen.width != width || Screen.height != height) {
        width = Screen.width;
        height = Screen.height;

        if (postProcessingCameraCommandBuffer != null) {
          postProcessingCamera.RemoveCommandBuffer(POST_PROC_EVENT, postProcessingCameraCommandBuffer);
        }

        if (commandBufferResources.ready) {
          postProcessingCameraCommandBuffer = CommandBufferBuilder.buildFullCommandBuffer(commandBufferResources, width, height, postProcessingCamera.stereoEnabled);
          postProcessingCamera.AddCommandBuffer(POST_PROC_EVENT, postProcessingCameraCommandBuffer);
        }
      }


      // absolute MESS
      // TODO: look at SteamVRFrustumAdjust in disassembly

      // could be suffering from this issue: https://github.com/ValveSoftware/openvr/issues/1052
      if (postProcessingCamera.stereoEnabled) {
        // postProcessingCamera.ResetStereoProjectionMatrices();
        // postProcessingCamera.ResetStereoViewMatrices();

        // TODO: maybe false here should be true cause we are rendering to a render texture... could also be mistake in shader cause of this
        Matrix4x4 leftProjection = Matrix4x4.identity;
        Matrix4x4 rightProjection = Matrix4x4.identity;

        leftProjection = postProcessingCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Left);
        rightProjection = postProcessingCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Right);

        // if (SuperController.singleton.isOpenVR) {
        //   leftProjection = SuperController.singleton.ViveCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Left);
        //   rightProjection = SuperController.singleton.ViveCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Right);
        // } else if (SuperController.singleton.isOVR) {
        //   leftProjection = SuperController.singleton.OVRCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Left);
        //   rightProjection = SuperController.singleton.OVRCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Right);
        // } else {
        //   SuperController.LogError("projection - yet another vr camera");
        // }

        // Matrix4x4[] inverseProjection = {
        //   GL.GetGPUProjectionMatrix(leftProjection, true).inverse,
        //   GL.GetGPUProjectionMatrix(rightProjection, true).inverse,
        // };

        // // apply the fix from https://forum.unity.com/threads/solved-reconstruct-world-position-from-depth-texture-in-single-pass-stereo-vr.480003/
        // inverseProjection[0][1, 1] *= - 1;
        // inverseProjection[1][1, 1] *= - 1;

        // // TODO: try this, ref https://docs.unity3d.com/Manual/SL-PlatformDifferences.html

        // Matrix4x4 leftView = Matrix4x4.identity;
        // Matrix4x4 rightView = Matrix4x4.identity;

        // if (SuperController.singleton.isOpenVR) {
        //   leftView = SuperController.singleton.ViveCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Left);
        //   rightView = SuperController.singleton.ViveCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Right);
        // } else if (SuperController.singleton.isOVR) {
        //   leftView = SuperController.singleton.OVRCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Left);
        //   rightView = SuperController.singleton.OVRCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Right);
        // } else {
        //   SuperController.LogError("view - yet another vr camera");
        // }

        // SuperController.LogMessage($"{leftView.GetRow(0)} {leftView.GetRow(1)} {leftView.GetRow(2)} {leftView.GetRow(3)}");

        // Matrix4x4[] inverseView = {
        //   (leftView).inverse,
        //   (rightView).inverse,
        // };

        var gpu_left = GL.GetGPUProjectionMatrix(leftProjection, true);
        var gpu_right = GL.GetGPUProjectionMatrix(rightProjection, true);

        Matrix4x4[] clipToWorld = {
          (gpu_left * postProcessingCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Left)).inverse,
          (gpu_right * postProcessingCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Right)).inverse,
        };

        // prepassMaterial.SetMatrixArray("_InverseView", inverseView);
        // prepassMaterial.SetMatrixArray("_InverseProjection", inverseProjection);
        commandBufferResources.prepassMaterial.SetMatrixArray("_ClipToWorld", clipToWorld);
      } else {
        // var viewMatrix = postProcessingCamera.worldToCameraMatrix.inverse;
        // var projectionMatrix = GL.GetGPUProjectionMatrix(postProcessingCamera.projectionMatrix, true).inverse;

        // projectionMatrix[1, 1] *= - 1;
        // projectionMatrix[1, 1] *= - 1;

        var ctw = (GL.GetGPUProjectionMatrix(postProcessingCamera.projectionMatrix, true) * postProcessingCamera.worldToCameraMatrix).inverse;

        // Matrix4x4[] inverseView = { viewMatrix, viewMatrix };
        // Matrix4x4[] inverseProjection = { projectionMatrix, projectionMatrix };
        Matrix4x4[] clipToWorld = { ctw, ctw };
      
        // prepassMaterial.SetMatrixArray("_InverseView", inverseView);
        // prepassMaterial.SetMatrixArray("_InverseProjection", inverseProjection);
        commandBufferResources.prepassMaterial.SetMatrixArray("_ClipToWorld", clipToWorld);
      }
    }
    
    private void OnDisable() {     
      // SuperController.LogMessage("EffectsCamera.OnDisable()");
      if (mainCamera != null)
        mainCamera.clearFlags = CameraClearFlags.SolidColor;
        mainCamera.cullingMask = DEFAULT_CULLING_MASK;

      postProcessingCamera.enabled = false;
    }

    private void OnDestroy() {
      // SuperController.LogMessage("EffectsCamera.OnDestroy()");
      // postProcessingCamera.RemoveCommandBuffer(PRE_PROC_EVENT, prepassCommandBuffer);
      postProcessingCamera.RemoveCommandBuffer(POST_PROC_EVENT, postProcessingCameraCommandBuffer);
      if (mainWindowCamera != null) mainWindowCamera.RemoveCommandBuffer(POST_PROC_EVENT, mainWindowCameraCommandBuffer);
      SuperController.singleton.screenshotCamera.RemoveCommandBuffer(POST_PROC_EVENT, screenshotCameraCommandBuffer);
      SuperController.singleton.hiResScreenshotCamera.RemoveCommandBuffer(POST_PROC_EVENT, hiResScreenshotCameraCommandBuffer);
      // mainCamera.RemoveCommandBuffer(CameraEvent.AfterImageEffects, finalCommandBuffer);

      // CustomPostProcessing customPostProcessing = getOrAddCustomPostProcessingComponent();
      // DestroyImmediate(customPostProcessing);
      DestroyImmediate(postProcessingCamera);

      GameObject cameraContainer = GameObject.Find("Moyashi_PostProcessingCamera");
      if (cameraContainer != null) DestroyImmediate(cameraContainer);      
    }
    
    private void OnPreCull() {
      postProcessingCamera.cullingMatrix = mainCamera.cullingMatrix;
    }

    private void recalculateSharpeningConsts() {
      float inputScreenX = postProcessingCamera.scaledPixelWidth;
      float inputScreenY = postProcessingCamera.scaledPixelHeight;

      float outputScreenX = postProcessingCamera.pixelWidth;
      float outputScreenY = postProcessingCamera.pixelHeight;

      var uniforms = SharedUniforms.singleton;

      uniforms.SharpeningConst0[0] = Util.f32asuint32(inputScreenX * (1.0f / outputScreenX)); 
      uniforms.SharpeningConst0[1] = Util.f32asuint32(inputScreenY * (1.0f / outputScreenY));
      uniforms.SharpeningConst0[2] = Util.f32asuint32(0.5f * inputScreenX * (1.0f / outputScreenX) - 0.5f);
      uniforms.SharpeningConst0[3] = Util.f32asuint32(0.5f * inputScreenY * (1.0f / outputScreenY) - 0.5f);

      float sharp = -(1.0f / (Mathf.Lerp(8.0f, 5.0f, uniforms.SharpeningScale)));
      Vector2 hSharp = new Vector2(sharp, 0.0f);

      uniforms.SharpeningConst1[0] = Util.f32asuint32(sharp);
      uniforms.SharpeningConst1[1] = Util.f32tof16(hSharp[0]) + (Util.f32tof16(hSharp[1]) << 16); // pack 2x16 bit float
      uniforms.SharpeningConst1[2] = Util.f32asuint32(8.0f * inputScreenX * (1.0f / outputScreenX));
      uniforms.SharpeningConst1[3] = 0.0f;
    }
  }
}